//
//  cu_kd_tree.h̃Tv
//
#include <stdio.h>
#include <limits.h>
#include <time.h>
#include <cutil.h> 
#include <cuda.h>
#include <cutil_inline.h>

#include "../cu_min_pq.h"
#include "../cu_kd_tree.h"


__global__ void gpu_run(kd_node_t *kd_nodes, fpoint_t *nodes, int *idx, int w, int h, int dim, int h1, int h2)
{

	int x = threadIdx.x + blockDim.x * blockIdx.x; 
	int y = threadIdx.y + blockDim.y * blockIdx.y;

	if(y >= h1 && y < h2&& x < w)
	{
		fpoint_t in;
		for(int i = 0;i < dim;i++)
			in.x[i] = nodes[x].x[i];

		min_pq pq;
		pq.n = 0;
		idx[x + y * w] = cu_kd_tree_search(&kd_nodes[0], &in, dim, 100, &pq, kd_nodes, nodes);
	}
}




int main(int argc, char** argv)
{
	int num, dim;
	int t0;
	int w, h;
	w = h = 256;
	int max_it = w * h;

	// ̗vfƎ
	num = 50000;
	dim = 128;

	// vf̃[N蓖
	fpoint_t   *nodes;
	nodes = new fpoint_t[num];	

	// vfKɐݒ
	srand(0);
	int i, j;
	for(i = 0;i < num;i++)
	{
		for(j = 0;j < dim;j++)
		{
			nodes[i].x[j] = rand()% 10000 / 10000.0;
		};// j
	};// i

	// kd؂̍\zBkd؂̗vf͓Kɑ߂ɊmۂĂB
	int kd_node_num = num * 2;
	kd_node_t *kd_nodes;
	kd_nodes = new kd_node_t[kd_node_num];	
	printf("build kd-tree...");
	kdtree_build(0, num, dim, kd_nodes, nodes);
	printf("done\n");

	fpoint_t node;
	min_pq  minpq;

   //foCX̏
    CUT_DEVICE_INIT(argc, argv);


	//************************************************************************************************
	// CPUō\zkd؂GPUփRs[
	kd_node_t *cu_kd_nodes;
	fpoint_t  *cu_nodes;
	cudaMalloc((void**)&cu_kd_nodes, kd_node_num * sizeof(kd_node_t));
	cudaMalloc((void**)&cu_nodes,            num * sizeof(fpoint_t));
	int *cu_idx, *idx;
	idx = new int[max_it];
	cudaMalloc((void**)&cu_idx, max_it * sizeof(int));
	// fotփ]
	cudaMemcpy(cu_kd_nodes,kd_nodes,  kd_node_num * sizeof(kd_node_t), cudaMemcpyHostToDevice );
	cudaMemcpy(cu_nodes,nodes,                num * sizeof(fpoint_t),  cudaMemcpyHostToDevice );
	//************************************************************************************************



	int dimx = 16;
	int dimy = 16;
	dim3 Grid(dimx, dimy);			//Block structure, threads/block limited by specific device
	dim3 Block(w/dimx, h/dimy);	    //Grid structure


	// Ts
	printf("iteration=%d, pts_num=%d, dim=%d\nNN-search...", max_it, num, dim);
	t0 = clock();

	int step = 8;
	int ndiv = h / step;

	for(int k = 0;k < ndiv;k++)
	{	
		gpu_run<<<Grid, Block>>>(cu_kd_nodes, cu_nodes, cu_idx, w, h, dim, k * step, (k + 1) * step);
		cudaThreadSynchronize();
	};

	printf("done\nCPU time = %d[msec]\n", clock() - t0);
	cudaMemcpy(idx, cu_idx, max_it * sizeof(int),  cudaMemcpyDeviceToHost );

	// vZʂ`FbN
	int c = 0;
	for(int i = 0;i < w*h;i++)
	{
		if(idx[i] != i%w)
			c++;
	};
	printf("fail num=%d\n", c);
	// 
	cudaFree(cu_kd_nodes);
	cudaFree(cu_nodes);
	cudaFree(cu_idx);

    // I
    CUT_EXIT(argc, argv);


	return 0;
};
